/*
 * libspe - A wrapper library to adapt the JSRE SPU usage model to SPUFS
 * Copyright (C) 2005 IBM Corp.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License,
 * or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 *  License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software Foundation,
 *   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <elf.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <sched.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/poll.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/syscall.h>

#include <sys/spe.h>
#include "libspe.h"
#include "spe.h"
#include "elf_loader.h"
#include "spe_exec.h"
#include "default_c99_handler.h"
#include "default_posix1_handler.h"


#define __PRINTF(fmt, args...) { fprintf(stderr,fmt , ## args); }
#ifdef DEBUG
#define DEBUG_PRINTF(fmt, args...) __PRINTF(fmt , ## args)
#else
#define DEBUG_PRINTF(fmt, args...)
#endif


/*
 * Helpers
 *
 * */

static struct group_list grp_list = {NULL,PTHREAD_MUTEX_INITIALIZER,0};

void env_check(void);
static int __env_spu_debug_start = 0;
static int __env_spu_info = 0;
static int check_env = 1;


/*Default SPE group options
 */
//static struct group_store default_group = {SCHED_OTHER, 0, 0, NULL, 0 };

/*Default SPE library call handlers for 21xx stop-and-signal.
*/
static void *handlers[] = {
	default_c99_handler, default_posix1_handler,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL };

/*
 * int spe_ldr[]:
 * SPE code that performs the actual parameter setting:
 */
static int spe_ldr[] = {
	0x30fff203,
	0x30fff404,
	0x30fff605,
	0x30fff806,
	0x30fff000,
	0x35000000,
	0x00000000,
	0x00000000
};

#if 0
/*
 * int dump_regs[]:
 * SPE code that saves register content to local store:
 */
static int dump_regs[] = {
        0x20a00000, 0x20a00201, 0x20a00402, 0x20a00603,
        0x20a00804, 0x20a00a05, 0x20a00c06, 0x20a00e07,
        0x20a01008, 0x20a01209, 0x20a0140a, 0x20a0160b,
        0x20a0180c, 0x20a01a0d, 0x20a01c0e, 0x20a01e0f,
        0x20a02010, 0x20a02211, 0x20a02412, 0x20a02613,
        0x20a02814, 0x20a02a15, 0x20a02c16, 0x20a02e17,
        0x20a03018, 0x20a03219, 0x20a0341a, 0x20a0361b,
        0x20a0381c, 0x20a03a1d, 0x20a03c1e, 0x20a03e1f,
        0x20a04020, 0x20a04221, 0x20a04422, 0x20a04623,
        0x20a04824, 0x20a04a25, 0x20a04c26, 0x20a04e27,
        0x20a05028, 0x20a05229, 0x20a0542a, 0x20a0562b,
        0x20a0582c, 0x20a05a2d, 0x20a05c2e, 0x20a05e2f,
        0x20a06030, 0x20a06231, 0x20a06432, 0x20a06633,
        0x20a06834, 0x20a06a35, 0x20a06c36, 0x20a06e37,
        0x20a07038, 0x20a07239, 0x20a0743a, 0x20a0763b,
        0x20a0783c, 0x20a07a3d, 0x20a07c3e, 0x20a07e3f,
        0x20a08040, 0x20a08241, 0x20a08442, 0x20a08643,
        0x20a08844, 0x20a08a45, 0x20a08c46, 0x20a08e47,
        0x20a09048, 0x20a09249, 0x20a0944a, 0x20a0964b,
        0x20a0984c, 0x20a09a4d, 0x20a09c4e, 0x20a09e4f,
        0x20a0a050, 0x20a0a251, 0x20a0a452, 0x20a0a653,
        0x20a0a854, 0x20a0aa55, 0x20a0ac56, 0x20a0ae57,
        0x20a0b058, 0x20a0b259, 0x20a0b45a, 0x20a0b65b,
        0x20a0b85c, 0x20a0ba5d, 0x20a0bc5e, 0x20a0be5f,
        0x20a0c060, 0x20a0c261, 0x20a0c462, 0x20a0c663,
        0x20a0c864, 0x20a0ca65, 0x20a0cc66, 0x20a0ce67,
        0x20a0d068, 0x20a0d269, 0x20a0d46a, 0x20a0d66b,
        0x20a0d86c, 0x20a0da6d, 0x20a0dc6e, 0x20a0de6f,
        0x20a0e070, 0x20a0e271, 0x20a0e472, 0x20a0e673,
        0x20a0e874, 0x20a0ea75, 0x20a0ec76, 0x20a0ee77,
        0x20a0f078, 0x20a0f279, 0x20a0f47a, 0x20a0f67b,
        0x20a0f87c, 0x20a0fa7d, 0x20a0fc7e, 0x20a0fe7f,
        0x00002000, 0x00000000, 0x00000000, 0x00000000
};

/*
 * int restore_regs[]:
 * SPE code that restores register content from local store:
 */
static int restore_regs[] = {
        0x30a00000, 0x30a00201, 0x30a00402, 0x30a00603,
        0x30a00804, 0x30a00a05, 0x30a00c06, 0x30a00e07,
        0x30a01008, 0x30a01209, 0x30a0140a, 0x30a0160b,
        0x30a0180c, 0x30a01a0d, 0x30a01c0e, 0x30a01e0f,
        0x30a02010, 0x30a02211, 0x30a02412, 0x30a02613,
        0x30a02814, 0x30a02a15, 0x30a02c16, 0x30a02e17,
        0x30a03018, 0x30a03219, 0x30a0341a, 0x30a0361b,
        0x30a0381c, 0x30a03a1d, 0x30a03c1e, 0x30a03e1f,
        0x30a04020, 0x30a04221, 0x30a04422, 0x30a04623,
        0x30a04824, 0x30a04a25, 0x30a04c26, 0x30a04e27,
        0x30a05028, 0x30a05229, 0x30a0542a, 0x30a0562b,
        0x30a0582c, 0x30a05a2d, 0x30a05c2e, 0x30a05e2f,
        0x30a06030, 0x30a06231, 0x30a06432, 0x30a06633,
        0x30a06834, 0x30a06a35, 0x30a06c36, 0x30a06e37,
        0x30a07038, 0x30a07239, 0x30a0743a, 0x30a0763b,
        0x30a0783c, 0x30a07a3d, 0x30a07c3e, 0x30a07e3f,
        0x30a08040, 0x30a08241, 0x30a08442, 0x30a08643,
        0x30a08844, 0x30a08a45, 0x30a08c46, 0x30a08e47,
        0x30a09048, 0x30a09249, 0x30a0944a, 0x30a0964b,
        0x30a0984c, 0x30a09a4d, 0x30a09c4e, 0x30a09e4f,
        0x30a0a050, 0x30a0a251, 0x30a0a452, 0x30a0a653,
        0x30a0a854, 0x30a0aa55, 0x30a0ac56, 0x30a0ae57,
        0x30a0b058, 0x30a0b259, 0x30a0b45a, 0x30a0b65b,
        0x30a0b85c, 0x30a0ba5d, 0x30a0bc5e, 0x30a0be5f,
        0x30a0c060, 0x30a0c261, 0x30a0c462, 0x30a0c663,
        0x30a0c864, 0x30a0ca65, 0x30a0cc66, 0x30a0ce67,
        0x30a0d068, 0x30a0d269, 0x30a0d46a, 0x30a0d66b,
        0x30a0d86c, 0x30a0da6d, 0x30a0dc6e, 0x30a0de6f,
        0x30a0e070, 0x30a0e271, 0x30a0e472, 0x30a0e673,
        0x30a0e874, 0x30a0ea75, 0x30a0ec76, 0x30a0ee77,
        0x30a0f078, 0x30a0f279, 0x30a0f47a, 0x30a0f67b,
        0x30a0f87c, 0x30a0fa7d, 0x30a0fc7e, 0x30a0fe7f,
        0x00002000, 0x00000000, 0x00000000, 0x00000000
};
#endif


void
env_check(void)
{
  if (getenv("SPU_DEBUG_START"))
    {
      __env_spu_debug_start = strtol ((const char *)getenv("SPU_DEBUG_START"),(char **)NULL, 0);
    }
  if (getenv("SPU_INFO"))
    {
      __env_spu_info = strtol ((const char *)getenv("SPU_INFO"),(char **)NULL, 0);
    }
  check_env = 0;
}


static int
spe_open_files(struct thread_store *spe, const char *pathname)
{
	char filename[256];
	int ret;
	
#ifdef NOSYSCALL
	sprintf(filename, "%s/run", pathname);
	spe->fd_run = open(filename, O_RDONLY);
#endif
	sprintf(filename, "%s/mbox", pathname);
	spe->fd_mbox = open(filename, O_RDONLY);

	sprintf(filename, "%s/mbox_stat", pathname);
	spe->fd_mbox_stat = open(filename, O_RDONLY);

	sprintf(filename, "%s/ibox", pathname);
	spe->fd_ibox = open(filename, O_RDONLY);

	sprintf(filename, "%s/ibox_stat", pathname);
	spe->fd_ibox_stat = open(filename, O_RDONLY);

	sprintf(filename, "%s/wbox", pathname);
	spe->fd_wbox = open(filename, O_WRONLY);

	sprintf(filename, "%s/wbox_stat", pathname);
	spe->fd_wbox_stat = open(filename, O_RDONLY);

	sprintf(filename, "%s/npc", pathname);
	spe->fd_npc = open(filename, O_RDWR);
		
	sprintf(filename, "%s/signal1", pathname);
	spe->fd_sig1 = open(filename, O_WRONLY);
		
	sprintf(filename, "%s/signal2", pathname);
	spe->fd_sig2 = open(filename, O_WRONLY);

	ret=pipe(spe->ev_pipe);
		
	if (spe->fd_mbox      < 0 || spe->fd_ibox      < 0 || spe->fd_wbox      < 0 ||
	    spe->fd_mbox_stat < 0 || spe->fd_ibox_stat < 0 || spe->fd_wbox_stat < 0 ||
	    spe->fd_npc       < 0 || spe->fd_sig1      < 0 || spe->fd_sig2      < 0 || 
	    ret ) {
		DEBUG_PRINTF("Could not open all SPE files.\n");
		close(spe->fd_mbox);
		close(spe->fd_mbox_stat);
		close(spe->fd_ibox);
		close(spe->fd_ibox_stat);
		close(spe->fd_wbox);
		close(spe->fd_wbox_stat);
		close(spe->fd_npc);
		close(spe->fd_sig1);
		close(spe->fd_sig2);
		close(spe->ev_pipe[0]);
		close(spe->ev_pipe[1]);
		return -1;
	}
	return 0;
}


/**
 * Low-level API
 * 
 */
int add_group_to_groups(struct group_store *gid)
{
	struct grpsElem *listElem;
	listElem = malloc(sizeof *listElem);

	if (!listElem)
	{
		errno=ENOMEM;
		return -errno;
	}
	
	pthread_mutex_lock(&grp_list.mutex);

	listElem->group=gid;
	listElem->next=grp_list.list_start;
	
	grp_list.list_start=listElem;
	grp_list.numListSize++;
		
	pthread_mutex_unlock(&grp_list.mutex);
	
	return 0;
	
}
int add_thread_to_group(struct group_store *gid, struct thread_store *thread)
{
	struct grpListElem *listElem;
	listElem = malloc(sizeof *listElem);

	if (!listElem)
	{
		errno=ENOMEM;
		return -errno;
	}

	pthread_mutex_lock(&grp_list.mutex);
	
	listElem->thread=thread;
	listElem->next=gid->grp_members;

	gid->grp_members=listElem;
	gid->numListSize++;
		
	pthread_mutex_unlock(&grp_list.mutex);
	
	return 0;
	
}

int remove_thread_from_group(struct group_store *gid, struct thread_store *thread)
{
	pthread_mutex_lock(&grp_list.mutex);
	
	if (gid->numListSize == 1)
	{	
		struct grpListElem *p=gid->grp_members;
		gid->numListSize--;
		free(p);
		gid->grp_members = NULL;
		pthread_mutex_unlock(&grp_list.mutex);
		return 0;
	}

	for (struct grpListElem *p=gid->grp_members; p->next!=NULL; p=p->next)
	{
		struct grpListElem *np=p->next;
		if (np->thread==thread)
		{
			p->next=np->next;
			gid->numListSize--;
			free(np);
			
			pthread_mutex_unlock(&grp_list.mutex);

			return 0;
		}
	}

	pthread_mutex_unlock(&grp_list.mutex);

	errno = ESRCH;
	return -ESRCH;
}

void *
spe_gid_setup(int policy, int priority, int use_events)
{
	struct group_store *group_store;

        group_store = calloc (1, sizeof *group_store);
	if (!group_store)
	{
		DEBUG_PRINTF ("Could not allocate group store\n");
		errno = ENOMEM;
	        return NULL;
	}
	
	group_store->policy = policy;
	group_store->priority = priority;
	group_store->use_events = use_events;
	group_store->grp_members = NULL;
	group_store->numListSize = 0;
	
	return group_store;
}

void *
spe_setup (spe_gid_t gid, spe_program_handle_t *handle, void *argp, void *envp, int flags)
{
	struct thread_store *thread_store;
	int rc, memfd, sigfd;
	addr64 argp64, envp64, tid64;
	char signame[256], memname[256], pathname[256];
	void *spe_ld_buf;
#ifdef SPE_NO_MMAP
	ssize_t count = 0, num = 0;
#endif
	struct spe_ld_info ld_info;
	struct spe_exec_params spe_params __attribute__ ((aligned (4096)));
	
	if(!gid)
	{
		gid=spe_gid_setup(SCHED_OTHER,0,0);
	}
	else
	{
		struct group_store *grp = gid;
		if (grp->numListSize == MAX_THREADS_PER_GROUP)
		{
		  	DEBUG_PRINTF ("Thread group has reached maximum number of members.\n");
		  	errno = ENOMEM;
		  	return NULL;
			
		}
	}

	/*Allocate thread structure */

	thread_store = calloc (1, sizeof *thread_store);
	if (!thread_store)
	  {
		  DEBUG_PRINTF ("Could not allocate thread store\n");
		  errno = ENOMEM;
		  return NULL;
	  }

	argp64.ull = (unsigned long long) (unsigned long) argp;
	envp64.ull = (unsigned long long) (unsigned long) envp;

	thread_store->group_id = gid;

	tid64.ull = (unsigned long long) (unsigned long) thread_store;
	
	/* Make the SPU Directory */

	sprintf (pathname, "/spu/spethread-%i-%lu",
		 getpid (), (unsigned long)thread_store);

	DEBUG_PRINTF ("create %s\n", pathname);

#ifdef NOSYSCALL
	rc = mkdir(pathname, S_IRUSR | S_IWUSR | S_IXUSR);
#else
	rc = spe_create(pathname, 0, S_IRUSR | S_IWUSR | S_IXUSR);
	thread_store->fd_run = rc;
#endif
	if (rc < 0)
	  {
		  DEBUG_PRINTF ("Could not create SPE %s\n", pathname);
		  return NULL;
	  }

	sprintf (memname, "%s/mem", pathname);

	/* Check SPE */

	memfd = open (memname, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR);
	if (memfd < 0)
	  {
		  DEBUG_PRINTF ("Could not open SPE mem file.\n");
		  return NULL;
	  }

	/* Prepare Loader */
#ifdef SPE_NO_MMAP	
	spe_ld_buf = malloc (LS_SIZE);
#else
	spe_ld_buf = mmap(0, LS_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, memfd, 0);
	thread_store->mem_mmap_base = spe_ld_buf;
	thread_store->fd_mem = memfd;
#endif
	if (!spe_ld_buf)
	  {
		  DEBUG_PRINTF ("Could not allocate SPE memory. \n");
		  errno = ENOMEM;
		  return NULL;
	  }

	memset(spe_ld_buf, 0, LS_SIZE);

	rc = load_spe_elf (handle, spe_ld_buf, &ld_info);
	if (rc != 0)
	  {
		  DEBUG_PRINTF ("Load SPE ELF failed..\n");
		  return NULL;
	  }

	thread_store->handle = *handle;

	/* Add SPE exec program */

	DEBUG_PRINTF ("Add exec prog dst:0x%04x size:0x%04x\n",
		      SPE_LDR_START, sizeof (spe_ldr));
	memcpy (spe_ld_buf + SPE_LDR_START, &spe_ldr, sizeof (spe_ldr));

	/* Add SPE exec parameters */

	spe_params.gpr0[0] = ld_info.entry;
	spe_params.gpr0[1] = 0;
	spe_params.gpr0[2] = 0;
	spe_params.gpr0[3] = 0;
	
	if (flags & SPE_USER_REGS) {
		/* When flags & SPE_USER_REGS is set, argp points
		 * to an array of 3x128b registers to be passed
		 * directly to the SPE program.
		 */
		memcpy(&spe_params.gpr3[0], argp, sizeof(unsigned int) * 12);
	} else {
		/* Regular parameter convention for spe_create_thread. */
		spe_params.gpr3[0] = tid64.ui[0];
		spe_params.gpr3[1] = tid64.ui[1];
		spe_params.gpr3[2] = 0;
		spe_params.gpr3[3] = 0;
	
		spe_params.gpr4[0] = argp64.ui[0];
		spe_params.gpr4[1] = argp64.ui[1];
		spe_params.gpr4[2] = 0;
		spe_params.gpr4[3] = 0;
	
		spe_params.gpr5[0] = envp64.ui[0];
		spe_params.gpr5[1] = envp64.ui[1];
		spe_params.gpr5[2] = 0;
		spe_params.gpr5[3] = 0;
	}

	spe_params.gpr6[0] = 0;
	spe_params.gpr6[1] = 0;
	spe_params.gpr6[2] = 0;
	spe_params.gpr6[3] = 0;

	DEBUG_PRINTF ("Add exec param dst:0x%04x size:0x%04x\n",
		      SPE_PARAM_START, sizeof (spe_params));
	memcpy (spe_ld_buf + SPE_PARAM_START, &spe_params,
		sizeof (spe_params));

#ifdef SPE_NO_MMAP
	/* Copy SPE image to SPUfs */
	do
	  {
		  num = write (memfd, spe_ld_buf + count, LS_SIZE - count);
		  if (num == -1)
		    {
			    DEBUG_PRINTF ("Transfer SPE ELF failed..\n");
			    return NULL;
		    }

		  count += num;
	  }
	while (count < LS_SIZE && num);
	close (memfd);

	/* Free the SPE Buffer */
	free (spe_ld_buf);
#endif

	rc = spe_open_files(thread_store, pathname);
	if (rc) {
		perror("open files\n");
		return NULL;
	}

	rc = add_thread_to_group(gid, thread_store);
        
	if ( flags & SPE_CFG_SIGNOTIFY1_OR )
        {
                char one[] = "1";
		int ignore;

                sprintf (signame, "%s/signal1_type", pathname);
                sigfd = open (signame, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR);
                if (sigfd < 0)
                {
                        DEBUG_PRINTF ("Could not open SPE mem file.\n");
                        return NULL;
                }
                ignore=write (sigfd, one, 1);
                close(sigfd);
        }
        if ( flags & SPE_CFG_SIGNOTIFY2_OR )
        {
                char one[] = "1";
		int ignore;
   
                sprintf (signame, "%s/signal2_type", pathname);
                sigfd = open (signame, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR);
                if (sigfd < 0)
                {
                        DEBUG_PRINTF ("Could not open SPE mem file.\n");
                        return NULL;
                } 
                ignore=write (sigfd, one, 1);
                close(sigfd); 
        }

	sync();

	return thread_store;
}

void
spe_cleanup (void *ptr)
{
	struct thread_store *thread_store = ptr;
	char pathname[256];

	sprintf (pathname, "/spu/spethread-%i-%lu",
		 getpid (), (unsigned long)thread_store);

	close (thread_store->fd_run);
	close (thread_store->fd_mbox);
	close (thread_store->fd_mbox_stat);
	close (thread_store->fd_ibox);
	close (thread_store->fd_ibox_stat);
	close (thread_store->fd_wbox);
	close (thread_store->fd_wbox_stat);
	close (thread_store->fd_npc);
	close (thread_store->fd_sig1);
	close (thread_store->fd_sig2);

#ifndef SPE_NO_MMAP
	munmap(thread_store->mem_mmap_base,LS_SIZE);
	close (thread_store->fd_mem);
#endif
	close (thread_store->ev_pipe[0]);
        close (thread_store->ev_pipe[1]);

#if defined(NOSYSCALL) && !defined(NODIRREMOVE)
	rmdir (pathname);
#endif
	remove_thread_from_group(thread_store->group_id, thread_store);

	free (thread_store);
}

unsigned int
set_npc (void *ptr, unsigned int npc)
{
	struct thread_store *thread_store = ptr;
	int ret;
	char npcText[9];

	sprintf(npcText,"0x%06x\n",npc);
	ret = write(thread_store->fd_npc, npcText, 9);
	if (ret < 0)
        {
                return -EINVAL;
        }
	return npc;
}

#ifdef NOSYSCALL
int
do_spe_run (void *ptr)
{
	struct thread_store *thread_store = ptr;
	int runfd;

	struct spufs_run_arg
	{
		unsigned int npc;	/* in/out: Next Program Counter */
		unsigned short code;	/* out:   SPU status */
		unsigned short status;
	};
	struct spufs_run_arg arg = { SPE_LDR_PROG_start, 0, 0 };
	int ret, rc;

        DEBUG_PRINTF ("do_spe_run\n");
	runfd=thread_store->fd_run;

	ret = set_npc(thread_store, arg.npc);
	if (ret < 0)
	{
		DEBUG_PRINTF ("Could not set NPC in NPC file.\n");
		return -EINVAL;
	}

	do {

		ret = read(runfd, &arg, 8);
                
		DEBUG_PRINTF ("SPE (npc=0x%06x) (code=0x%04x) (status=0x%04x).\n",
			      arg.npc, arg.code, arg.status);
                if ( ret < 0) {
                          DEBUG_PRINTF ("Could not read() on SPE run file.\n");
			  return -EINVAL;
                } else if ((arg.code&0xff00)==0x2100) {
			int callnum = arg.code&0xff;
			
			DEBUG_PRINTF ("SPE library call: ");
			
			if (!handlers[callnum])
			{
				DEBUG_PRINTF ("Unsupported SPE library call signal: 21%i.",callnum);
				break;
			}
			else
			{
				int (*handler)(void *, unsigned int);
				handler=handlers[callnum];
				
				rc = handler(thread_store->mem_mmap_base, arg.npc);
				if (rc) 
				{
			        	DEBUG_PRINTF ("SPE library call unsupported.\n");
				        return -ENOSYS;
			        }
			}
		}
		else if ((arg.code&0xf000)==0x1000)
		{
			pthread_mutex_lock(&thread_store->event_lock);
			pthread_cond_init(&thread_store->event_deliver, NULL);
			
			if (((struct group_store*)thread_store->group_id)->use_events)
			{
				int callnum = arg.code;
				//
				// Event - driven 21xx hanlers
				//
				
				thread_store->event = SPE_EVENT_STOP;
				thread_store->ev_data = callnum;
				//
				//Todo Error handling
				//
				write(thread_store->ev_pipe[1], &thread_store->event, 4);

				//Show that we're waiting
				thread_store->event_pending = 1;
				
				// Wait for a signal.
				pthread_cond_wait(&thread_store->event_deliver, &thread_store->event_lock);
				
				thread_store->event_pending = 0;
				
				pthread_mutex_unlock(&thread_store->event_lock);
			}
			else
			{
				pthread_mutex_unlock(&thread_store->event_lock);
		        	
				DEBUG_PRINTF ("SPE user events not enabled.\n");
			        return -ENOSYS;
		        }
		}

        } while ((arg.code & 0xff00) == 0x2100 || (arg.code & 0xf000) == 0x1000);

	DEBUG_PRINTF ("SPE thread result: %08x:%04x:%04x\n", arg.npc,
		      arg.code, arg.status);
	
	/* Save status & npc */	
	thread_store->npc = arg.npc;
	thread_store->ret_status = arg.status;

	//
	//Event code
	//
	if (((struct group_store*)thread_store->group_id)->use_events)
	{
		thread_store->event = SPE_EVENT_THREAD_EXIT;
		thread_store->ev_data = arg.code;
		//
		//Todo Error handling
		//
		write(thread_store->ev_pipe[1], &thread_store->event, 4);
	}

	return arg.code;
}
#else
int
do_spe_run (void *ptr)
{
	struct thread_store *thread_store = ptr;
	int runfd;

	unsigned int npc;
	unsigned int code;
	unsigned int status;

	npc = SPE_LDR_PROG_start;

	int ret, rc;

        DEBUG_PRINTF ("do_spe_run\n");
	runfd = thread_store->fd_run;

	do {
		ret = spe_run(runfd, &npc, &status);
		code = ret >> 16;
                DEBUG_PRINTF ("SPE (npc=0x%06x) (code=0x%04x) (ret=0x%04x).\n",
			      npc, code, ret & 0xffff);

                if (ret < 0) {
                          DEBUG_PRINTF ("Could not run on SPE: %s.\n",
				strerror(errno));
			  return -errno;
                } else if ((code&0xff00)==0x2100) {
			int callnum = code&0xff;
			
			DEBUG_PRINTF ("SPE library call: ");
			
			if (!handlers[callnum])
			{
				DEBUG_PRINTF ("Unsupported SPE library call signal: 21%i.",callnum);
				break;
			}
			else
			{
				int (*handler)(void *, unsigned int);
				handler=handlers[callnum];
				
				rc = handler(thread_store->mem_mmap_base, npc);
				if (rc) 
				{
			        	DEBUG_PRINTF ("SPE library call unsupported.\n");
				        return -ENOSYS;
			        }
				else
				{
					npc += 4;
				}
			}

		}
		else if (code < 0x2000)
		{
			pthread_mutex_lock(&thread_store->event_lock);
			pthread_cond_init(&thread_store->event_deliver, NULL);
			
			if (((struct group_store*)thread_store->group_id)->use_events)
			{
				int callnum = code;
				//
				// Event - driven 21xx hanlers
				//
				
				thread_store->event = SPE_EVENT_STOP;
				thread_store->ev_data = callnum;
				//
				//Todo Error handling
				//
				write(thread_store->ev_pipe[1], &thread_store->event, 4);
				
				//Show that we're waiting
				thread_store->event_pending = 1;
				
				// Wait for a signal.
				pthread_cond_wait(&thread_store->event_deliver, &thread_store->event_lock);

				thread_store->event_pending = 0;
				
				pthread_mutex_unlock(&thread_store->event_lock);

			}
			else
			{

				pthread_mutex_unlock(&thread_store->event_lock);
				
				DEBUG_PRINTF ("SPE user events not enabled.\n");
			        return -ENOSYS;
			}
		}

        } while ((code & 0xff00) == 0x2100 || (code < 0x2000));

	DEBUG_PRINTF ("SPE thread result: %08x:%04x\n", npc, code);
	
	/* Save status & npc */	
	thread_store->npc = npc;
	thread_store->ret_status = ret;
	
	//
	//Event code
	//
	if (((struct group_store*)thread_store->group_id)->use_events)
	{
		thread_store->event = SPE_EVENT_THREAD_EXIT;
		thread_store->ev_data = code;
		//
		//Todo Error handling
		//
		write(thread_store->ev_pipe[1], &thread_store->event, 4);
	}

	return code;
}
#endif

struct image_handle {
	spe_program_handle_t speh;
	unsigned int map_size;
};

spe_program_handle_t
*spe_open_image(const char *filename)
{
	/* allocate an extra integer in the spe handle to keep the mapped size information */
	struct image_handle *ret;
	int binfd = -1, f_stat;
	struct stat statbuf;
	size_t ps = getpagesize ();

	ret = malloc(sizeof(struct image_handle));
	if (!ret)
		return NULL;

	ret->speh.handle_size = sizeof(spe_program_handle_t);
	ret->speh.toe_shadow = NULL;

	binfd = open(filename, O_RDONLY);
	if (binfd < 0)
		goto ret_err;

	f_stat = fstat(binfd, &statbuf);
	if (f_stat < 0)
		goto ret_err;

	/* now store the size at the extra allocated space */
	ret->map_size = (statbuf.st_size + ps - 1) & ~(ps - 1);

	ret->speh.elf_image = mmap(NULL, ret->map_size, PROT_READ , MAP_SHARED, binfd, 0);
	if (ret->speh.elf_image == MAP_FAILED)
		goto ret_err;

	/* ok */
	close(binfd);
	return (spe_program_handle_t *)ret;

	/* err & cleanup */
ret_err:
	if (binfd >= 0)
		close(binfd);

	free(ret);
	return NULL;
}

int
spe_close_image(spe_program_handle_t *handle)
{
	void *elf_start = handle->elf_image;
	unsigned int prog_size = ((struct image_handle*)handle)->map_size;

	free(handle);

	return munmap(elf_start, prog_size);
}


void
register_handler(void * handler, unsigned int callnum ) 
{
	handlers[callnum]=handler;
}


/**
 * Library API
 * 
 */

spe_gid_t
spe_create_group (int policy, int priority, int use_events)
{
	struct group_store *group_store;

	DEBUG_PRINTF ("spu_create_group(0x%x, 0x%x, 0x%x)\n",
			                  policy, priority, use_events);
	
	group_store=spe_gid_setup(policy, priority, use_events);
	DEBUG_PRINTF("gid is %p\n", group_store);

	return group_store;
}

int
spe_group_max (spe_gid_t spe_gid)
{
	DEBUG_PRINTF ("spu_group_max(0x%x)\n", spe_gid);

	return MAX_THREADS_PER_GROUP;
}


static void *
spe_thread (void *ptr)
{
	/* If the SPU_INFO (or SPU_DEBUG_START) environment variable is set,
	   output a message to stderr telling the user how to attach a debugger
	   to the new SPE thread.  */
	if (__env_spu_debug_start || __env_spu_info)
	  {
		int tid = syscall (__NR_gettid);
		fprintf (stderr, "Starting SPE thread %p, to attach debugger use: spu-gdb -p %d\n",
			 ptr, tid);

		/* In the case of SPU_DEBUG_START, actually wait until the user *has*
		   attached a debugger to this thread.  This is done here by doing an
		   sigwait on the empty set, which will return with EINTR after the
		   debugger has attached.  */
		if (__env_spu_debug_start)
		  {
			sigset_t set;
			int sig;
			sigemptyset (&set);
			sigwait (&set, &sig);
		  }
	  }

	return (void *) (unsigned long) do_spe_run (ptr);
}

speid_t
spe_create_thread (spe_gid_t gid, spe_program_handle_t *handle, void *argp, void *envp, unsigned long mask, int flags)
{
	struct thread_store *thread_store;
	int rc;

	DEBUG_PRINTF ("spe_create_thread(%p, %p, %p, %p, 0x%x, 0x%x)\n",
		      gid, handle, argp, envp, mask, flags);

	if (check_env)
	  env_check();

	thread_store = spe_setup (gid, handle, argp, envp, flags);
	if (!thread_store)
	  {
		  DEBUG_PRINTF ("spe_setup failed\n");
		  errno=EINVAL;
		  return NULL;
	  }
	
	rc = pthread_create (&thread_store->spe_thread, NULL,
			     spe_thread, thread_store);
	if (rc)
	  {
		  perror ("pthread_create");
		  return NULL;
	  }

	return thread_store;
}

static inline int spe_wait_status(unsigned int spe_status_R)
{
	unsigned int rc, term = 0;

	rc = (spe_status_R & 0xffff0000) >> 16;
	if (spe_status_R & 0x4) {
		/* SPE halted. */
		term = SIGABRT;
	} else if (spe_status_R & 0x20) {
		/* SPE invalid instruction. */
		term = SIGILL;
	} else if (spe_status_R & 0x40) {
		/* SPE invalid ch. */
		term = SIGILL;
	} else if ((rc < 0x2000) || (rc > 0x20ff)) {
		/* Treat these as invalid stop codes. */
		term = SIGILL;
	} else if (spe_status_R & 0x1) {
		/* Thread killed for undetermined reason. */
		term = SIGKILL;
	}
	/* Return status that can be evaluated with WIFEXITED(), etc. */
	return ((rc & 0xff) << 8) | (term & 0x7f);
}

int 
spe_wait(speid_t speid, int *status, int options)
{
	struct thread_store *thread_store = speid;
	void *pthread_rc;
	int rc;

	DEBUG_PRINTF("spu_wait(0x%x, %p, 0x%x)\n", speid, status, options);

	if (options) {
		DEBUG_PRINTF("  'options' not supported.\n");
		errno = EINVAL;
		return -1;
	}
	rc = pthread_join(thread_store->spe_thread, &pthread_rc);
	if (rc) {
		DEBUG_PRINTF("  pthread_join failed, errno=%d\n", errno);
		return -1;
	}
	if (status) {
		*status = spe_wait_status(thread_store->ret_status);
	}
	spe_cleanup(thread_store);

	DEBUG_PRINTF("Thread ended.\n");
	return rc;
}

#if 0
int
spe_wait_save (speid_t speid, struct spe_ctx *spe_context)
{
	struct thread_store *thread_store = speid;
	void *pthread_rc;
	int rc, ret;

	struct spufs_run_arg
	{
		unsigned int npc;       /* inout: Next Program Counter */
	        unsigned short code;    /* out:   SPU status */
	        unsigned short status;
        };
        struct spufs_run_arg arg = { 0, 0, 0 };

	rc = pthread_join (thread_store->spe_thread, &pthread_rc);
	if (rc) {
		DEBUG_PRINTF("  pthread_join failed, errno=%d\n", errno);
		return -1;
	}
	
#ifdef SPE_NO_MMAP
        spe_cleanup (thread_store);
	
	DEBUG_PRINTF ("spe_wait_save(): This function is only available on mmapped spes.\n");
	
	return -1;
#else
	if (!spe_context->ls)
	{
		DEBUG_PRINTF ("spe_wait_save(): No LS to copy to.\n");
		return -1;
	}

	/* Save local store */
	memcpy (spe_context->ls, thread_store->mem_mmap_base, LS_SIZE);

	/* Save results */
	spe_context->npc = thread_store->npc;
	spe_context->status = thread_store->ret_status;
	
	/* Upload dump register code */
	memcpy (thread_store->mem_mmap_base + 0x1000, &dump_regs, sizeof(dump_regs));
	
	/* Run the register dumper */
	set_npc(thread_store, 0x1000);
	ret = read(thread_store->fd_run, &arg, 8);

	/* Save the registers */
	memcpy (&spe_context->reg, thread_store->mem_mmap_base + 0x10000, sizeof(spe_context->reg));

#endif

	spe_cleanup (thread_store);

	DEBUG_PRINTF ("Thread ended.\n");
	return rc;
}

#endif

int
spe_kill (speid_t speid, int sig)
{
	struct thread_store *thread_store = speid;
	int rc;

	if (sig == SIGCONT)
	{
		int stopped = 0;
		
		pthread_mutex_lock(&thread_store->event_lock);
		stopped = thread_store->event_pending;
		pthread_mutex_unlock(&thread_store->event_lock);
		if (stopped)
		{
			pthread_cond_signal(&thread_store->event_deliver);
		}
		return 0;
	}

	rc = pthread_kill (thread_store->spe_thread, sig);

	return rc;
}

int
spe_get_event(struct spe_event *pevents, int nevents, int timeout)
{
	int i;
	int ret_events = 0;
	int numSPEsToPoll = 0;
	int setupSPEs =0;
	int pollRet = 0;

	struct pollfd *SPEfds;

	struct poll_helper {
		int event;
		struct thread_store *thread;
		int type;
	};

	struct poll_helper *phelper;
	
	pthread_mutex_lock(&grp_list.mutex);

	for (i=0 ; i < nevents ; i++)
	{
		int j;
		struct group_store *group = pevents[i].gid;
		struct grpListElem *elem = group->grp_members;
		
		// Clear output fields.
		pevents[i].speid = NULL;
		pevents[i].revents=0;
		pevents[i].data=0;

		DEBUG_PRINTF("spe_get_event():\n");
		DEBUG_PRINTF("  using group  : %p\n", group);
		DEBUG_PRINTF("  with members : %i\n", group->numListSize);

		for ( j=0 ; j < group->numListSize; j++)
		{
			struct thread_store *thread = elem->thread;
		
			DEBUG_PRINTF("  scan member  : %p\n", thread);
			
			// Scan for present events in the thread structures of SPEs
			if (thread->event != 0 && thread->event & pevents[i].events)
			{
				int pipeval,ret;
			
				DEBUG_PRINTF("  has event !  : 0x%04x\n", thread->event);
				
				//Found an event we're looking for.
				ret_events++;

				//Fill out return struct.
				pevents[i].revents = thread->event;
				pevents[i].speid   = thread;
				pevents[i].data    = thread->ev_data;

				//Empty the pipe
				ret=read(thread->ev_pipe[0], &pipeval, 4);
				thread->ev_data = 0;
				thread->event = 0;

				pthread_mutex_unlock(&grp_list.mutex);
				return ret_events;
			}
			
			//Decide on what fd's to poll on
			if (pevents[i].events & (SPE_EVENT_MAILBOX))
			{
				numSPEsToPoll++;
			}
			if (pevents[i].events & (SPE_EVENT_STOP | SPE_EVENT_TAG_GROUP | SPE_EVENT_DMA_ALIGNMENT | 
						SPE_EVENT_SPE_ERROR | SPE_EVENT_SPE_DATA_SEGMENT | 
						SPE_EVENT_SPE_DATA_STORAGE | SPE_EVENT_SPE_TRAPPED | SPE_EVENT_THREAD_EXIT ))
			{
				numSPEsToPoll++;
			
			}
			elem=elem->next;
		}
	}
	
	if(numSPEsToPoll == 0)
	{
		pthread_mutex_unlock(&grp_list.mutex);
		errno=EINVAL;
		return -1;
	}

	if (ret_events > 0)
	{
		//printf("P1\n");
		
		DEBUG_PRINTF("  returning !  : 0xi\n", ret_events);

		pthread_mutex_unlock(&grp_list.mutex);
		
		return ret_events;
	}
	
	DEBUG_PRINTF("  number of fd : %i\n", numSPEsToPoll);

	SPEfds=malloc (numSPEsToPoll * sizeof(struct pollfd));
	phelper=malloc (numSPEsToPoll * sizeof(struct poll_helper));

	// Set up all necessary fds to poll on and remeber what they are for.
	for (i=0 ; i < nevents ; i++)
	{
		int j;
		struct group_store *group = pevents[i].gid;
		struct grpListElem *elem = group->grp_members;

		for ( j=0 ; j < group->numListSize; j++)
		{
			struct thread_store *thread = elem->thread;
			
			if (pevents[i].events & (SPE_EVENT_MAILBOX))
			{
				SPEfds[setupSPEs].fd=thread->fd_ibox;
				SPEfds[setupSPEs].events=POLLIN;
			
				phelper[setupSPEs].event=i;
				phelper[setupSPEs].thread=thread;
				phelper[setupSPEs].type=1;
				//printf("1\n");
				setupSPEs++;
			}
			if (pevents[i].events & (SPE_EVENT_STOP | SPE_EVENT_TAG_GROUP | SPE_EVENT_DMA_ALIGNMENT | 
						SPE_EVENT_SPE_ERROR | SPE_EVENT_SPE_DATA_SEGMENT | 
						SPE_EVENT_SPE_DATA_STORAGE | SPE_EVENT_SPE_TRAPPED | SPE_EVENT_THREAD_EXIT ))
			{
				SPEfds[setupSPEs].fd=thread->ev_pipe[0];
				SPEfds[setupSPEs].events=POLLIN;
				
				phelper[setupSPEs].event=i;
				phelper[setupSPEs].thread=thread;
				phelper[setupSPEs].type=2;
				//printf("2\n");
				
				setupSPEs++;
			
			}
			elem=elem->next;
		}
	}
	
	if (setupSPEs != numSPEsToPoll)
	{
		DEBUG_PRINTF("ERROR:Thread number mismatch.");
		
		free(SPEfds);
		free(phelper);
		
		pthread_mutex_unlock(&grp_list.mutex);
		errno = EFAULT;
		return -1;
	}

	pthread_mutex_unlock(&grp_list.mutex);

	pollRet = poll(SPEfds, setupSPEs, timeout);
	
	// Timeout.
	if (pollRet == 0)
	{
		free(SPEfds);
		free(phelper);
		return 0;
	}
	
	
	for (i=0 ; i < setupSPEs ; i++ )
	{
		if (SPEfds[i].revents != 0)
		{
			if (phelper[i].type==1)
			{
				// Read ibox data if present
				int rc,data;

				rc = read(SPEfds[i].fd, &data, 4);
				if (rc == 4)
				{
					pevents[phelper[i].event].data = data;
					pevents[phelper[i].event].speid = phelper[i].thread;
					pevents[phelper[i].event].revents = SPE_EVENT_MAILBOX;
				}

			}
			else
			{
                               	// Read pipe data if present
                               	int rc,data;

                               	rc = read(SPEfds[i].fd, &data, 4);
                               	//pevents[phelper[i].event].revents = pevents[phelper[i].event].revents | data;
                               	pevents[phelper[i].event].revents = data;
				pevents[phelper[i].event].speid = phelper[i].thread;
			       	pevents[phelper[i].event].data = phelper[i].thread->ev_data;

			       	phelper[i].thread->ev_data = 0;
			       	phelper[i].thread->event = 0;

			}
		}
	}
	
	free(SPEfds);
	free(phelper);
	//printf("P2\n");
	return pollRet;
}

int
spe_get_threads(spe_gid_t gid, speid_t *spe_ids)
{
	struct group_store *group = gid;
	int i;
	
	if (!spe_ids)
	{
		return group->numListSize;
	}
	else
	{
		struct grpListElem *elem = group->grp_members;
		
		for(i=0; i < group->numListSize ; i++)
		{
			spe_ids[i] = elem->thread;
			elem=elem->next;
			
		}
	}

	return i;
}

/*
int
spe_mbox_read(speid_t speid, unsigned int *val)
{
	struct thread_store *thread_store = speid;
	int rc;

	rc = read(thread_store->fd_mbox, val, 4);
	if (rc == 4)
		rc = 0;

	return rc;
}

int
spe_ibox_read(speid_t speid, unsigned int *val)
{
	struct thread_store *thread_store = speid;
	int rc;

	rc = read(thread_store->fd_ibox, val, 4);
	if (rc == 4)
		rc = 0;

	return rc;
}

int
spe_wbox_write(speid_t speid, unsigned int val)
{
	struct thread_store *thread_store = speid;
	int rc;

	rc = write(thread_store->fd_wbox, &val, 4);
	if (rc == 4)
		rc = 0;

	return rc;
}
*/

void
*spe_get_ls (speid_t speid)
{
	struct thread_store *thread_store = speid;
	
#ifdef SPE_NO_MMAP
	errno = EFAULT;
	return -1;
#else
	return thread_store->mem_mmap_base;
#endif
}

int spe_get_priority(spe_gid_t gid)
{
	struct group_store *group_store = gid;
	
	return group_store->priority;
}

int spe_set_priority(spe_gid_t gid, int priority)
{
	struct group_store *group_store = gid;
	
	group_store->priority=priority;

	/*int pthread_setschedparam(pthread_t target_thread,  int  policy,  const struct sched_param *param);*/
	return 0;
}

int spe_get_policy(spe_gid_t gid)
{
	struct group_store *group_store = gid;
	
	return group_store->policy;
}

spe_gid_t
spe_get_group (speid_t speid)
{
	struct thread_store *thread_store = speid;
	return thread_store->group_id;
}

void
*spe_get_ps (speid_t speid)
{
	return speid;
}

int
spe_get_affinity( speid_t speid, unsigned long *mask)
{
	printf("spe_get_affinity: not implemented in this release.\n");

	errno=ENOSYS;
	return -1;
}

int
spe_set_affinity(speid_t speid, unsigned long mask)
{
	printf("spe_set_affinity: not implemented in this release.\n");

	errno=ENOSYS;
	return -1;
}

int
spe_get_context(speid_t speid, struct spe_ucontext *uc)
{
	printf("spe_get_context: not implemented in this release.\n");

	errno=ENOSYS;
	return -1;
}

int
spe_set_context(speid_t speid, struct spe_ucontext *uc)
{
	printf("spe_set_context: not implemented in this release.\n");

	errno=ENOSYS;
	return -1;
}

/*
 * mfc.h direct call-ins
 *
 * */

int spe_write_in_mbox (speid_t spu_ps_addr ,unsigned int data)
{
	struct thread_store *thread_store = spu_ps_addr;
	int rc;
	
	rc = write(thread_store->fd_wbox, &data, 4);
	if (rc == 4)
		rc = 0;

	return rc;

}

int spe_stat_in_mbox (speid_t spu_ps_addr)
{
	struct thread_store *thread_store = spu_ps_addr;
	int rc, ret;

	rc = read(thread_store->fd_wbox_stat, &ret, 4);
	if (rc != 4)
		ret = -1;

	return ret;

}

unsigned int spe_read_out_mbox (speid_t spu_ps_addr)
{
        struct thread_store *thread_store = spu_ps_addr;
        int rc;
	unsigned int ret;

        rc = read(thread_store->fd_mbox, &ret, 4);
        if (rc != 4)
	        ret = -1;

        return ret;

}


int spe_stat_out_mbox (speid_t spu_ps_addr)
{
	struct thread_store *thread_store = spu_ps_addr;
        int rc, ret;

        rc = read(thread_store->fd_mbox_stat, &ret, 4);
        if (rc != 4)
                ret = -1;

        return ret;

}


int spe_stat_out_intr_mbox (speid_t spu_ps_addr)
{
        struct thread_store *thread_store = spu_ps_addr;
        int rc, ret;

        rc = read(thread_store->fd_ibox_stat, &ret, 4);
        if (rc != 4)
                ret = -1;

        return ret;

}


int spe_write_signal (speid_t spu_ps_addr, unsigned int signal_reg, unsigned int data)
{
	struct thread_store *thread_store = spu_ps_addr;
	int rc;

	if (signal_reg == SPE_SIG_NOTIFY_REG_1)
		rc = write(thread_store->fd_sig1, &data, 4);
	else if (signal_reg == SPE_SIG_NOTIFY_REG_2)
		rc = write(thread_store->fd_sig2, &data, 4);
	else
		return -1;
	
	if (rc == 4)
		rc = 0;

	return rc;
}

