/*
 * libspe - A wrapper library to adapt the JSRE SPU usage model to SPUFS 
 * Copyright (C) 2005 IBM Corp. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by 
 * the Free Software Foundation; either version 2.1 of the License, 
 * or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful, but 
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 *  License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License 
 *   along with this library; if not, write to the Free Software Foundation, 
 *   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#ifndef _spe_h_
#define _spe_h_

/* Definitions
 */

struct thread_store
{
	/* Note: The debugger accesses this data structure, and assumes it
	   starts out with an spe_program_handle_t identifying the SPE
	   executable running within this SPE thread.  Everything below
	   is private to libspe.  */
	spe_program_handle_t handle;

        pthread_t 	spe_thread;
	pthread_cond_t 	event_deliver;
	pthread_mutex_t event_lock;
        int 		fd_mem, fd_run, fd_mbox, fd_mbox_stat, fd_ibox, 
			fd_ibox_stat, fd_wbox, fd_wbox_stat, fd_npc, fd_sig1, 
			fd_sig2,
		 	policy, priority, affinity_mask, spe_flags,
			npc, ret_status,event, ev_pipe[2];
        spe_gid_t 	group_id;
        void 		*mem_mmap_base;
	unsigned long 	ev_data;
	int 		event_pending;
};

/*Group members list
 */

struct grpListElem
{
        struct grpListElem *next;
        struct thread_store *thread;
};

struct group_store
{
        int 		policy, priority, use_events;
        struct 		grpListElem *grp_members;
        int 		numListSize;
};
/*Groups List
 */
struct grpsElem
{
        struct grpsElem 	*next;
        struct group_store 	*group;
};

struct group_list
{
        struct grpsElem *list_start;
	pthread_mutex_t	mutex;
        int 		numListSize;
};

struct spe_ctx {
	unsigned int npc;
	unsigned int status;
	unsigned long reg[128][4];		// size error for LP64
	void *ls;
};




/* Low-level SPE execution API.
 */

extern int add_group_to_groups(struct group_store *gid);
extern int add_thread_to_group(struct group_store *gid, struct thread_store *thread);
extern int remove_thread_from_group(struct group_store *gid, struct thread_store *thread);

extern void *spe_gid_setup(int policy, int priority, int use_events);
extern void *spe_setup (spe_gid_t gid, spe_program_handle_t *handle, void *argp, void *envp, int flags);
extern void spe_cleanup (void *spe);
extern int do_spe_run (void *spe);
extern unsigned int set_npc (void *spe, unsigned int npc);

extern void register_handler(void * handler, unsigned int callnum );

/*
 * For testing purposes only 
 *
extern int spe_wait_save (speid_t speid, struct spe_ctx *spe_context);
extern int spe_mbox_read(speid_t speid, unsigned int *val);
extern int spe_ibox_read(speid_t speid, unsigned int *val);
extern int spe_wbox_write(speid_t speid, unsigned int val);
*/

#endif
