/*
 * libspe - A wrapper library to adapt the JSRE SPU usage model to SPUFS
 * Copyright (C) 2005 IBM Corp.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License,
 * or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 *  License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software Foundation,
 *   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <libspe.h>

int main(int argc, char *argv[])
{
	spe_program_handle_t *binary[16];
	speid_t spe_thread[16];
	spe_gid_t spe_group;
	int result = 0;
	int i, j, max_spe, threads, status, ret;

	static int test_result __attribute__ ((aligned(128)));

	/* Basic Test */

	printf("\nStep 1: spe_create_group() ");
	
	spe_group = spe_create_group(SCHED_OTHER, 0, 0);
	if (!spe_group) 
	{
		printf("error: create_group.\n");
		result++;
	}
	printf(" OK.\n");
	printf("  spe_group=%p.\n",spe_group);
 
	printf("\nStep 2: spe_group_max() returns:");
	
	max_spe = spe_group_max(spe_group);
	printf("%i\n", max_spe);

	printf("\nStep 3: spe_get_threads(spe_gid_t,NULL) returns:");
	threads = spe_get_threads(spe_group,NULL);
	printf("%i ", threads);
	if(threads)
	{
		printf(" ERROR.\n");
		result++;
	}
	else
		printf(" OK.\n");
	
	printf("\nStep 4a: spe_open_image(\"spe-test-pause\") ");
	binary[0] = spe_open_image("spe-test-pause");
	if (!binary[0])
	{
		printf(" ERROR.\n");
		exit(2);
	}
	printf(" OK.\n");


	printf("\nStep 4b: spe_open_image(\"spe-test-pause\") (again) ");
	binary[1] = spe_open_image("spe-test-pause");
	if (!binary[1])
	{
		printf(" ERROR.\n");
		exit(2);
	}
	printf(" OK.\n");


	printf("\nStep 5a: spe_create_thread() ");
	spe_thread[0] = spe_create_thread(spe_group, binary[0], &test_result, NULL, 0, 0);
	if (!spe_thread[0])
	{
		printf(" ERROR.\n");
		exit(2);
	}
	else
	{
		printf("OK.\n");
		printf("  Returned speid=%p\n",spe_thread[0]);
	}
	
	usleep(500000);

	printf("\nStep 5b: spe_create_thread() (again) ");
	spe_thread[1] = spe_create_thread(spe_group, binary[1], &test_result, NULL, 0, 0);
	if (!spe_thread[1])
	{
		printf(" ERROR.\n");
		exit(2);
	}
	else
	{
		printf("OK.\n");
		printf("  Returned speid=%p\n",spe_thread[1]);
	}

	printf("\nStep 6: spe_get_threads(spe_gid_t,NULL) returns:");
	threads = spe_get_threads(spe_group,NULL);
	printf("%i ", threads);
	if (threads != 2)
	{
		printf(" ERROR.\n");
		result++;
	}
	else
	{
		speid_t speids[16];
		int i;
		
		printf(" OK.\n");
		
		threads=spe_get_threads(spe_group,speids);
		for (i=0 ; i<threads ; i++)
		{
			printf("  Thread[%i] = %p\n",i,speids[i]);
		}
		
	}

	printf("\nStep 7a: spe_wait() ");
	ret = spe_wait(spe_thread[0], &status, 0);
	if(!ret && status== 0)
	{
		speid_t speids[16];
		int i;
		
		printf("OK.\n");
		printf("  Thread returned status=%i\n",status);
		printf("  Remaining threads:\n");
		
		threads=spe_get_threads(spe_group,speids);
		for (i=0 ; i<threads ; i++)
		{
			printf("  Thread[%i] = %p\n",i,speids[i]);
		}

	}
	else
	{
		printf("ERROR.\n");
		result++;
	
	}

	printf("\nStep 7b: spe_wait() ");
	ret = spe_wait(spe_thread[1], &status, 0);
	if(!ret && status== 0)
	{
		speid_t speids[16];
		int i;
		
		printf("OK.\n");
		printf("  Thread returned status=%i\n",status);
		printf("  Remaining threads:\n");
		
		threads=spe_get_threads(spe_group,speids);
		for (i=0 ; i<threads ; i++)
		{
			printf("  Thread[%i] = %p\n",i,speids[i]);
		}

	}
	else
	{
		printf("ERROR.\n");
		result++;
	
	}
	
	printf("\nStep 8: spe_get_threads(spe_gid_t,NULL) returns:");
	threads = spe_get_threads(spe_group,NULL);
	printf("%i ", threads);
	if(threads)
	{
		printf(" ERROR.\n");
		result++;
	}
	else
		printf(" OK.\n");

	spe_close_image(binary[0]);
	spe_close_image(binary[1]);

	printf("\nStep 9: Load tests\n");

	printf("  a.) lots of threads seqentially.");
	
	binary[0] = spe_open_image("spe-test-start-stop");

	for (i=0; i< 1024; i++)
	{
		spe_thread[0] = spe_create_thread(spe_group, binary[1], 
						  &test_result, NULL, 0, 0);

		if (!spe_thread[0])
               	{
			printf("\n    failed to create load thread %i.",i);
			result++;
			break;
		}
		spe_wait(spe_thread[0], &status, 0);
	}
	if (i==1024)
		printf(" OK.\n");

	printf("  b.) lots of threads seqentially and parallel in groups.");
	
	binary[0] = spe_open_image("spe-test-start-stop");

	for (i=0; i< 1024; i++)
	{
		for (j=0; j< 8; j++)
		{
			spe_thread[j] = spe_create_thread(spe_group, binary[1], 
						  &test_result, NULL, 0, 0);

			if (!spe_thread[j])
               		{
				printf("\n    failed to create load thread %i,%i.",i,j);
				result++;
				break;
			}
		}
		for (j=0; j< 8; j++)
		{
			spe_wait(spe_thread[j], &status, 0);
		}
	}
	if (i==1024)
		printf(" OK.\n");

	printf("\n\n\nTEST RESULT:.\n");
	printf("-------------\n");
	if (!result)
	{
		printf("All tests completed ok.\n");
	}
	else
	{
		printf("Errors were encountered - check your installation.\n");
	}

	return 0;
}
